#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <cmath>
#include <fstream>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include <boost/accumulators/accumulators.hpp>
#include <boost/accumulators/statistics/covariance.hpp>
#include <boost/accumulators/statistics/mean.hpp>
#include <boost/accumulators/statistics/stats.hpp>
#include <boost/accumulators/statistics/variance.hpp>
#include <boost/accumulators/statistics/variates/covariate.hpp>
#include <boost/format.hpp>
#include "GetPot"

// Print usage information
template<typename charT, typename traits>
void print_help(std::basic_ostream<charT, traits>& out,
                const std::string application)
{
    using namespace std;
    out << endl
        << "Compute the {x,y,z} mean, variance, and cross-correlations" << endl
        << "for a time-varying field." << endl
        << endl
        << "Usage:" << endl
        << "\t" << application << " [file]" << endl
        << endl
        << "Input file format:" << endl
        << "\tThe input file should contain four whitespace-separated" << endl
        << "\tcolumns.  Column 1 is the time index, and columns 2-4" << endl
        << "\tare the x, y, and z coordinate component, respectively." << endl
        << "\tOmitting [file] causes the program to process standard input." << endl
        << endl
        << "Output format:" << endl
        << "\tThe output contains ten columns: " << endl
        << "\t 1: time, copied from the original input" << endl
        << "\t 2: mean of x" << endl
        << "\t 3: mean of y" << endl
        << "\t 4: mean of z" << endl
        << "\t 5: variance of x" << endl
        << "\t 6: variance of y" << endl
        << "\t 7: variance of z" << endl
        << "\t 8: covariance of x and y" << endl
        << "\t 9: covariance of x and z" << endl
        << "\t10: covariance of y and z" << endl
        << endl;
}

// Report errors to cerr
void error(const std::string p, const std::string p2 = "") {
    using namespace std;
    cerr << p << ' ' << p2 << endl;
    std::exit(1);
}

// Read an input streams and compute the relevant statistics
template<typename charT, typename traits>
int process(std::basic_ostream<charT, traits>& out,
            std::basic_istream<charT, traits>& in) {

    using namespace boost::accumulators;
    accumulator_set<double, stats<tag::mean, 
                                  tag::variance, 
                                  tag::covariance<double, tag::covariate1> > > x_acc;
    accumulator_set<double, stats<tag::mean, 
                                  tag::variance, 
                                  tag::covariance<double, tag::covariate1> > > y_acc;
    accumulator_set<double, stats<tag::mean, 
                                  tag::variance, 
                                  tag::covariance<double, tag::covariate1> > > z_acc;

    double t, x, y, z;
    std::basic_string<charT, traits> line;
    // TODO Simplify formatting
    boost::format fmt(" %016e   %016e   %016e   %016e   %016e   %016e   %016e   %016e   %016e   %016e");
    while (in) {
        if (!std::getline(in, line)) break; // Read columns per line to
        std::istringstream in(line);       //    be a little defensive
        in >> t >> x >> y >> z;

        // Compute running statistics
        // TODO Can I use two separate covariates on x_acc here?
        x_acc(x, covariate1 = y);
        y_acc(y, covariate1 = z);
        z_acc(z, covariate1 = x);

        // Output running statistics
        out << fmt % t
                   % mean(x_acc)
                   % mean(y_acc)
                   % mean(z_acc)
                   % variance(x_acc)
                   % variance(y_acc)
                   % variance(z_acc)
                   % covariance(x_acc)
                   % covariance(z_acc)
                   % covariance(y_acc)
            << std::endl;
    } 

    return 0;
}

// Driver
int main( int argc, char **argv ) {
    GetPot cl(argc, argv);
    std::vector<std::string> files = cl.nominus_vector();

    if ( cl.search(2, "--help", "-h") ) {
        print_help(std::cout, cl[0]);
        exit(0);
    } else if ( files.size() > 1 ) {
        print_help(std::cerr, cl[0]);
        exit(1);
    }

    // TODO Incorporate http://www.gotw.ca/gotw/048.htm
    // TODO Do it for real this time
    int retval = 0;
    if (files.size() == 0) {
        retval = process(std::cout, std::cin);  // Process standard input
    } else {
        std::ifstream in(files[0].c_str());
        if (!in) error("Cannot open input file:", files[0]);

        retval = process(std::cout, in);
    }

    return retval;
}
